//
//  Class Calibrate: Provides methods for calibrating ASQ power spectra
//  into units of physical displacement (nanometers or strain).
//    version:  4.3 (Oct. 27 2003)   
//     author:  Kevin C. Schlaufman (kcs149@psu.edu)
//              Patrick J. Sutton (psutton@ligo.caltech.edu)
//
///////////////////////////////////////////////////////////////////////////
#ifndef Calibrate_HH
#define Calibrate_HH


#ifndef __CINT__
#include "FSpectrum.hh"
#include "LineFilter.hh"
#endif
/*
#include <math.h>//marc
#include <fstream.h>//marc
#include <iostream.h>//marc
*/


/** The Calibrate class is a helper class for SenseMonitor.  
  * It contains information required to convert FSpectrum 
  * data from AS_Q counts to physical strain.
  * @memo Class for calibrating ASQ FSpectrum data.
  */
class Calibrate
{
public:
        /** Create a calibrate object based on measured open-loop gain (H) 
          * and sensing function (C).  H, C data are read from specified 
          * files and linearly interpolated to specified frequencies.
	  * See documentation for formatting of these files.
	  * DYNAMIC
	  * @memo Constructor using open-loop gain and sensing function.
	  * @param olg_file File containing open-loop gain data (H). 
	  * @param sf_file File containing sensing function data (C). 
	  * @param channel Channel data taken from (eg: L1:LSC-AS_Q).
	  * @param Fmin Lowest frequency of data to be calibrated.
	  * @param Fstep Frequency spacing of data to be calibrated.
	  * @param npoint Number of data points to be calibrated.
	  */
  Calibrate(const char *olg_file, const char *sf_file, // const char *CalPSD_file,//-marc
		const char *channel, double Fmin, double Fstep, int npoint);

	//  Format of the static calibration file:
	//  --------------------------------------
	//  It is assumed that the first two rows of the calibration file 
        //  contain headers, and that the remaining rows of the file are in  
        //  the format 
        // 
        //    freq   modulus   modulus_error   phase   phase_error
        // 
        //  where modulus is the modulus of the transfer function 
        //  ETMX,Y to AS_Q [units (AS_Q counts)/(ETMX,Y counts)]. 
	// 
        //  :KLUDGE: I think need only first three columns - CHECK.
	// 
	//  It is assumed that the frequency is  
	//  monotonically increasing or decreasing.  Only the frequency and
        //  modulus are used, but there must be at least a third column for
        //  the file to be read properly.  Each row must not be more that 
        //  80 characters wide.
	//  It extracts the list of 
	//  frequencies and transfer function modulus values into the data 
        //  members double *freq, *module in ascending order of frequency.
	//  The DC pendulum transfer function value 'ETMX' is also 
        //  initialized according to the channel name using hardwired values.

        /** Create a Calibrate object based on a static calibration 
	  * for the ETM->ASQ transfer function.  This transfer function 
	  * is interpolated from data read from the specified file.
	  * See documentation for formatting of this file. 
	  * STATIC
	  * @memo Constructor using a static calibration.
	  * @param cal_file File containing data for the transfer function ETM->ASQ.
	  * @param channel Channel data taken from (eg: L1:LSC-AS_Q).
          */
	Calibrate(const char *cal_file, const char *channel);

        /** Destroy a calibrate object.
	  * DYNAMIC, STATIC
	  * @memo Destructor.
	  */
	~Calibrate();

	/** Calibrate the specified FSpectrum using a static calibration.
	  * This method extracts from the specified FSpectrum that frequency
	  * range for which static calibration data is available and calibrates
	  * it, converting the units from (AS_Q counts)^2/Hz to (nm)^2/Hz.  
	  * (This method reads the calibration data from the private data member
	  * FSpectrum *TransFunc.)
	  * STATIC
	  * @memo Calibrate the specified FSpectrum with a static calibration.
	  * @param Int_PSD Fspectrum to be calibrated.
	  */
	void ApplyStatic(FSpectrum& Int_PSD);

	/** Calibrate the specified FSpectrum using a static calibration.
	  * This method extracts from the specified FSpectrum the desired
	  * frequency range, uses linear interpolation to produce static
	  * calibration data for these frequencies, and applies the it,
	  * converting the units from (AS_Q counts)^2/Hz to (nm)^2/Hz.  
	  * (For each call this method re-computes the calibration data at 
	  * the required frequencies using linear interpolation.) 
	  * STATIC
	  * @memo Calibrate the specified FSpectrum over the specified frequency range.
	  * @param Int_PSD Fspectrum to be calibrated.
	  * @param low_freq Lowest frequency to be retained.
	  * @param high_freq Highest frequency to be retained.
	  */ 
	void ApplyStatic(FSpectrum& Int_PSD, double low_freq, double high_freq);

	/** Calibrate the specified FSpectrum using a dynamic calibration.
	  * This method extracts from the specified FSpectrum that frequency
	  * range for which dynamic calibration data is available and calibrates
	  * it, converting the units from (AS_Q counts)^2/Hz to (nm)^2/Hz.  
	  * (This method requires that the ComputeResponseFunction() have been 
	  * called first.)  
	  * DYNAMIC
	  * @memo Calibrate the specified FSpectrum with a dynamic calibration.
	  * @param Int_PSD Fspectrum to be calibrated.
	  */
        void ApplyDynamic(FSpectrum& Int_PSD);

	/** Measure the amplitude of a calibration line using S. Klimenko's 
	  * LineFilter class.  
	  * DYNAMIC, STATIC
	  * @memo Measure the amplitude of a line using a LineFilter object.
	  * @param ts Pointer to TSeries containing data with the line.
	  * @param Ref_Line Pointer to a LineFilter object for the line.
	  */
	float getAmplitude(TSeries *ts, LineFilter *Ref_Line);

	/** Measure the amplitude of a line using a simple 
	  * PSD-based estimate.  The procedure is
          *      1. Read in TSeries (preferably >=60 sec).
          *      2. Compute PSD with Hanning window, 10 averages.
          *      3. Line power = power in bin containing line freq, 
          *         + power in neighboring bins (3 bins total).
          *         - noise power in counted bins,
          *           estimated as average power in nearby bins.
          *      4. Line ampl = [ 2*dF*Line power ]^(0.5)
          *         where dF = n_windows/T = frequency resolution.
	  * DYNAMIC, STATIC
	  * @memo Make PSD-based estimate of line amplitude.
	  * @param ts Pointer to TSeries containing data with the line.
	  * @param freq Frequency of line.
	  */
        float getAmplitude(TSeries *ts, double Freq);

	/** Get ratio of current amplitude of calibration line to its initial 
	  * amplitude when the monitor was started (returns 1 when first called).
	  * Used as estimate of change in gain of IFO.
	  * STATIC
	  * @memo Get estimate of gain change of IFO.
	  * @param ts Pointer to TSeries of ASQ data.
	  * @param Ref_Line Pointer to LineFilter object for calibration line.
	  * @param lowerbound Minimum allowed amplitude.
	  * @param upperbound Maximum allowed amplitude.
	  * @param closeness Multiple of (maximum-minimum) amplitude 
          *        which line variance must be less than.
	  */
	float getGain(TSeries *ts, LineFilter *Ref_Line,
	 	      const double lowerbound, const double upperbound,
		      const double closeness);

	/** Return status of calibration-line amplitude.
	  * 0 - Gain is nominal.
	  * 1 - Gain is abnormally low (amplitude < lowerbound).
	  * 2 - Gain is abnormally high (amplitude > upperbound).
	  * 3 - Gain is varying wildly (mean absolute difference 
	  *     between last 5 consecutive amplitude measurements 
	  *     greater than closeness*(upperbound-lowerbound). 
	  * 4 - Gain cannot be computed (reported when regaining lock).
	  * STATIC
	  * @memo Return status of calibration-line amplitude.
	  */ 
	int getGainStatus(void);

	/** Compute modulus of static ETM->ASQ transfer function and store 
	  * in private data member FSpectrum* TransFunc.  This method uses 
          * linear interpolation to compute the transfer function at the 
	  * indicated frequencies.
	  * STATIC
	  * @memo Compute and store modulus of ETM->ASQ transfer function.
	  * @param f_min Minimum frequency at which to compute transfer function.
	  * @param f_step Frequency resolution at which to compute transfer function.
	  * @param nstep Number of frequency bins at which to compute transfer function.
	  */ 
	void InitializeTransferFunction(double f_min, double f_step, int nstep);

        /** Return minimum frequency for which calibration data is available.
	  * DYNAMIC, STATIC
	  * @memo Return minimum frequency for which calibration data is available.
	  */ 
        double getLowFreq(void);

        /** Return maximum frequency for which calibration data is available.
	  * DYNAMIC, STATIC
	  * @memo Return maximum frequency for which calibration data is available.
	  */ 
        double getHighFreq(void);

	/** Compute response function strain(nm)->ASQ.
	  * This method computes and stores the detector response function 
	  * using the open loop gain and sensing function.  This transfer 
	  * function converts from nm->ASQ counts.  This transfer function
	  * can be applied to an FSpectrum using the ApplyDynamic() method.
	  * The calibration parameters alpha and beta must already be 
	  * initialized.
	  * DYNAMIC
	  * @memo Compute response function strain(nm)->ASQ.
	  */ 
        void ComputeResponseFunction(void);

	/** Return the current value of alpha.
	  * DYNAMIC
	  * @memo Return the current value of alpha.
	  */ 
        double Calibrate::getAlpha(void);

	/** Return the current value of beta. 
	  * DYNAMIC
	  * @memo Return the current value of beta.
	  */ 
        double Calibrate::getBeta(void);

	/** Return the current value of gain. 
	  * DYNAMIC
	  * @memo Return the current value of gain.
	  */ 
        double Calibrate::getGain(void);

	/** Set alpha to the specified value.
	  * DYNAMIC
	  * @memo Set alpha to the specified value.
	  */ 
        void Calibrate::setAlpha(double value);

	/** Set beta to the specified value.
	  * DYNAMIC
	  * @memo Set beta to the specified value.
	  */ 
        void Calibrate::setBeta(double value);

	/** Set gain to the specified value.
	  * DYNAMIC
	  * @memo Set gain to the specified value.
	  */ 
        void Calibrate::setGain(double value);

	/** Compute a new value for alpha using the gain ratio of a calibration
	  * line at a specified frequency.  (The frequency is required for
	  * determining the value of the open-loop gain.) 
	  * DYNAMIC
	  * @memo Compute a new value for alpha.
	  * @param frequency  Frequency of the calibration line.
	  * @param ratio Ratio of the measure calibration-line amplitude in ASQ
	  *        to the expected value.
	  */ 
        void Calibrate::ComputeAlpha(double frequency, double ratio);

	/** Dump the open-loop gain and sensing function to the specified 
	  * output stream.  This method dumps every 60th interpolated data 
	  * point to the specified output stream.  It's useful for keeping 
	  * a record of what calibrations were used.  Header and tail 
	  * information is also dumped.
	  * DYNAMIC
	  * @memo Dump the open-loop gain and sensing function to the specified 
	  *       output stream.
	  * @param Output stream to dump the data to.
	  */ 
        void Calibrate::DumpDynamicCalibration(std::ostream& out);

	/** Dump the response function to the specified 
	  * output stream.  This method dumps every 60th interpolated data 
	  * point to the specified output stream.  Header and tail 
	  * information is also dumped.
	  * DYNAMIC
	  * @memo Dump the response function to the specified output stream.
	  * @param Output stream to dump the data to.
	  */ 
        void Calibrate::DumpResponseFunction(std::ostream& out);

	/** Returns true if the calibration is physically acceptable. 
	  * This method checks that the value of alpha is within 
	  * acceptable limits.
	  * DYNAMIC
	  * @memo Returns true if calibration is physically acceptable.
	  */ 
        bool Calibrate::IsCalibrationGood(void);

        //  The rest of the methods are for reading and manipulating data 
        //  from text files and have nothing in particular to do with calibration.

        /** Read data files containing three columns of numeric data and load  
          * data into the specified double arrays (memory for which must be 
          * assigned BEFORE calling this method; use, for example, 
          * Calibrate::GetDataFileLength(const char *file)).  Successive rows 
          * with repeated first element are ignored, and elements are arranged 
          * in increasing order by first column.  The method assumes the data 
	  * is monotonically increasing or decreasing in the data file.  
	  * Data in additional columns past the third are ignored.
	  * DYNAMIC, STATIC
	  * @memo Load data from file into double arrays.
	  * @param file Name of data file.
	  * @param numb_entry Number of rows in the file.
	  * @param column1 Pointer to array to hold data from the first column of the file.
	  * @param column2 Pointer to array to hold data from the second column of the file.
	  * @param column3 Pointer to array to hold data from the third column of the file.
	  */ 
        void Calibrate::ReadDataFile(const char *file, int numb_entry, double *column1, double *column2, double *column3);
  
  void Calibrate::ReadDataFile(const char *file, int numb_entry);// marc       
        /** Read a data file containing three columns of data and return the 
          * number of rows, ignoring successive rows with repeated first element. 
          * Assumes the first element of the first row is not 1.0.
	  * DYNAMIC, STATIC
	  * @memo Returns number of rows in the specified data file.
	  * @param file Name of the data file.
	  */ 
	int Calibrate::GetDataFileLength(const char *file);

        /** Using double arrays x, y of same length, linearly interpolate to calculate 
          * Y values at specified X. 
	  * DYNAMIC, STATIC
	  * @memo Linear interpolate (x,y) data to values Y at X.
	  * @param X Pointer to array of points at which to interpolate (x,y).
	  * @param Y Pointer to array to hold data output from interpolation.
	  * @param N Number of points of data in arrays X, Y.
	  * @param x Pointer to array holding data to be interpolated.
	  * @param y Pointer to array holding data to be interpolated.
	  * @param n Number of points of data in arrays x, y.
	  */ 
        void Calibrate::LinearInterpolation(double *X, double *Y, int N, double *x, double *y, int n);


//:KLUDGE: New cal code.

	/** This function updates the value of alpha stored in the Calibrate
	  * object.   UpdateBeta() should be called before calling this
	  * method, since alpha depends on beta.
	  * DYNAMIC
	  * @memo  Update calibration parameter alpha.
	  * @param ts_asq  Pointer to TSeries containing ASQ channel. 
	  * @param ts_exc  Pointer to TSeries containing EXC channel. 
	  * @param ampl_asq  Reference calibration-line amplitude in ASQ. 
	  * @param ampl_exc  Reference calibration-line amplitude in EXC. 
	  * @param freq  Frequency of calibration line. 
	  */ 
	void Calibrate::UpdateAlpha(TSeries* ts_asq, TSeries* ts_exc, double ampl_asq, double ampl_exc, double freq);


	/** This function updates the value of beta stored in the Calibrate
	  * object.  The new value is computed as the product of the averages 
	  * of the supplied TSeries data objects, divided by the reference
	  * value.  This method should be called before calling UpdateAlpha().
	  * DYNAMIC
	  * @memo  Update calibration parameter beta.
	  * @param ts_darm  Pointer to array of TSeries pointers containing DARM channels.
	  * @param num_darm_channels  Number of DARM channels in ts_darm.
	  * @param darm0  Reference DARM value.
	  */ 
	void Calibrate::UpdateBeta(TSeries** ts_darm, int num_darm_channels, double darm0);


//:END:KLUDGE:


private:

	//  True if computed alpha value is physically acceptable.
	bool isalphagood;

	//  boolean true if using dynamic calibration.
	bool dyncal;
	
	//  Current gain.
	double gain;
	
	//  Calibration parameters.
	double alpha;
	double beta;

	//  Real and imaginary parts of H at calibration-line freq.
	double reh;
	double imh;

	//  Common frequencies at which calibration data is stored (Npoint = number of freq).
	double *f;
        int Npoint;

        //  Response Function modulus and phase.
        double *modRF;
	double *phRF;

        //  Open-Loop Gain Function.
        double *modH;
	double *phH;

        //  Sensing Function.
        double *modC;
	double *phC;

	//  IFO arm length.
	float arm_length;

        //  Minimum and maximum frequencies for which calibration data is available.
        double Freq_min, Freq_max;

        //  Minimum, maximum frequencies for which calibration is applied
        //  (the intersection of [Freq_min,Freq_max] and the frequency range 
        //  of the AS_Q data, which is typically decimated to [0,4096]Hz).
        double freq_min, freq_max;

        //  Frequency spacing at which calibration data is applied.
        double freq_step;

        /////////////////////////////////////////////////////////////////////////
        //  The following data members are used by the STATIC calibration code 
        //  only.  None of these are used  by the dynamic calibration code.
        //  They are segregated for convenience.
        /////////////////////////////////////////////////////////////////////////

	//  Dummy line used to step through calibration file.
	char line[80];	

	//  Used to determine if any frequency steps are repeated.
        double next_entry, last_entry;

	//  Pointers to arrays which contain static calibration data.
	double *freq, *module;
	
	//  ETMX factors for each IFO:
        //  o December 28 2001: ETMX / ETMY calibrations were
	//       H1: 1.4/1.6,     H2: 1.33/1.35,   L1: 2.63/2.48
	//  o August 26 2002 (S1 run): ETMX / ETMY calibrations were
	//       H1: 1.05/1.20,   H2: 1.33/1.35,   L1: approx 2.6 (R. Adhikari)
	//  Most conservative is to use higher of ETMX, ETMY.  For S1 use
	//       H1: 1.20         H2: 1.35,        L1: 2.63
        //  Calibrate decides which nm/ETMX_count factor to use based on which 
        //  channel is being read (H1 channels use H1 ETMX, H2 channels use H2 
        //  ETMX, etc).
	double ETMX;
	double H1_ETMX;
	double H2_ETMX;
	double L1_ETMX;

	//  Pendulum frequency
	double F_PEND;

	//  Transfer function value at calibration-line frequency.
	double ASQ_per_ADC;

	//  Used to track the sum of the absolute values of the differences
	//  between the last RECENT number of calculated amplitudes.  If this
	//  number grows too large, then the calibration line is no good and
	//  it is no longer used to adjust the gain.  See also float var.
	double CLOSENESS;
	
	//  User defined parameters that determine on which interval the
	//  amplitude of the calibration line can be expected.  They are used to
	//  prevent the spikes in the calibration line that occur when it is 
	//  initialized or terminated from affecting calculated gain.
	double UPPERBOUND;
	double LOWERBOUND;

	//  Initial measured amplitude of calibration line; to be used as 
	//  reference for adjusting estimated range of IFO.
	float initial_amplitude;
	
	//  Used to track the sum of the absolute values of the differences
	//  between the last RECENT number of calculated amplitudes.  If this
	//  number grows too large, then the calibration line is no good and
	//  it is no longer used to adjust the gain.  See also double CLOSENESS.
	float var;

	//  Points to array of length RECENT which holds the last RECENT number
	//  of calculated amplitudes.
	float *last_few;

        //  FSpectrum holding modulus of Transfer_Function(AS_Q/ETMX,Y):
	FSpectrum* TransFunc;

	//  Length of freq and module arrays to be determined by constructor.
	int length;

	//  Keeps track of which element of last_few is most current
	int step;

	//  Status of calibration line amplitude:  0 nominal, 1 low (crashed),
        //  2 (spiked), 3 (wildly varying), 4 (re-initializing after lost lock
	//  or monitor start).  Used to pass warnings back to the calling
	//  program.
	int gain_status;

	//  Number of recent calculated amplitudes tracked.
	int RECENT;

	//  Holds end time of previous TSeries objects passed to getGain.
	//  If the the start time of the current TSeries is not equal to 
	//  prev_end_time (such as happens if lock was lost) then the 
	//  reference amplitude of the calibration line is reset.
	Time prev_end_time;
};


Calibrate::Calibrate(const char *olg_file, const char *sf_file,  //const char *CalPSD_file, // <- <- marc 
const char *channel, double Fmin, double Fstep, int npoint):isalphagood(true),dyncal(true),gain(0.0),alpha(0.0),beta(0.0),reh(0.0),imh(0.0)
{
/*
	Procedure:  
	  1) open olg file, read data, interpolate to desired frequencies.
	  2) open sf file, read data, interpolate to desired frequencies.
	  3) Combine olg & sf freq-by-freq and store in double arrays (later FSeries or FSpectrum).
	  4) Write new Apply method to combine with AS_Q data.
	Notes: 
	  o Don't want to store full olg and sf data as these files 
	    can be O(25MB) each, any why redo linear interpolation if 
            your frequency bins don't change?
	  o Frequency range specified in the arguments must be a subset 
            of the range covered by the calibration data, otherwise the 
            monitor will crash.  FIX OR PUT WARNING IN DOCUMENTS.
	To Do:
	  1) Check units (meters, strain, etc).
	  2) Need reference value of cal-line ampl in AS_Q counts.
	  3) Need to read reference values of cal-line ampl and beta from config file as 
	     part of -dyncal option.
	  4) Change -calfile option to -staticcal.
	  5) Make new option -dyncal olgfile sffile nom_ampl beta 
*/

	/*
	 * Assigned:
	 * Freq_min, Freq_max = min, max freq for which cal data is available.
	 * Npoint = number of frequencies calibration data is interpolated and stored at.
	 * float arrays f, modC, phC, modH, phH assigned and filled.
	 * float arrays modRF, phRF assigned but not filled.
	 * KLUDGE:  f[] is not used here.  Is it needed?
	 */ 
	//---------- Assign arm length in nanometers based on the channel name.
        //           This is used to change the response function from ASQ/strain 
        //           to ASQ/nm.  It's a bit whacko, but Range.hh assumes the 
        //           diferential arm length is in nm rather than strain.  It's 
        //           also convenient for avoiding over/underflow problems with 
        //           number storage.
	if ((channel[0] == 'H') && (channel[1] == '1')) {
            arm_length = 4.0e12;
        } else if ((channel[0] == 'H') && (channel[1] == '2')) {
            arm_length = 2.0e12;
        } else if ((channel[0] == 'L') && (channel[1] == '1')) {
            arm_length = 4.0e12;
        } else {
            std::cerr << "SenseMonitor ERROR:  Channel " << channel[0] 
		      << channel[1] << " not recognised.  Exiting.\n";
            std::cout << "SenseMonitor ERROR:  Channel " << channel[0] 
		      << channel[1] << " not recognised.  Exiting.\n";
	    exit(399);
	}


	//----- Read and interpolate data from calibration files.
	//---------- Create temporary variables.
	int temp_Npoint;
        double *temp_f;
        double *temp_mod;
        double *temp_ph;
	float *temp_array; //--marc did it
	//---------- Read files and determine frequency range they cover.
	//---------- Read data from Sensing Function file.
        temp_Npoint = GetDataFileLength(sf_file);  
        temp_f = new double[temp_Npoint];
        temp_mod = new double[temp_Npoint];
        temp_ph = new double[temp_Npoint];
        ReadDataFile(sf_file,temp_Npoint,temp_f,temp_mod,temp_ph);
	Freq_min = temp_f[0];
        Freq_max = temp_f[temp_Npoint-1];
        delete[] temp_f;
        delete[] temp_mod;
        delete[] temp_ph;
	/* /--marc start
	temp_Npoint = GetDataFileLength(CalPSD_file);
	temp_array = new float[temp_Npoint];
	// marc ReadDataFile(CalPSD_file, temp_Npoint);  
        */
	//-----marc end
	//---------- Read data from Open-Loop Gain file.
        temp_Npoint = GetDataFileLength(olg_file);  
        temp_f = new double[temp_Npoint];
        temp_mod = new double[temp_Npoint];
        temp_ph = new double[temp_Npoint];
        ReadDataFile(olg_file,temp_Npoint,temp_f,temp_mod,temp_ph);
	if (temp_f[0] > Freq_min) Freq_min = temp_f[0];
	if (temp_f[temp_Npoint-1] < Freq_max) Freq_max = temp_f[temp_Npoint-1];
        //---------- Extract subset of desired frequency range which 
	//           is covered by calibration data.  This determines
	//           how much storage we need.  (We don't assign storage 
	//           until we check the range covered by the sensing 
	//           data as well.) 
        double Fmax = Fmin + (npoint-1)*Fstep;
	//---------- Adjust low frequency.  Use <= instead of < for safety.
	int count; 
	count = 0;
	while (Fmin + count*Fstep <= Freq_min) ++count; 
	Fmin += count*Fstep;
	npoint -= count;
	//---------- Adjust high frequency.  Use >= instead of > for safety.
	count = 0;
	while (Fmax - count*Fstep >= Freq_max) ++count; 
	Fmax -= count*Fstep;
	npoint -= count;
	//---------- Allocate storage for calibration data. 
	Npoint = npoint;
	//---------- Frequencies 'f'.
        f = new double[Npoint];
        for (int i=0; i<Npoint; ++i) { 
            f[i] = Fmin + i*Fstep;
	}
        //:KLUDGE:  Make sure these don't conflict with freq_min, etc. from 
	//          older code.
	// (stored as doubles)
	freq_min = Fmin;
	freq_max = Fmax;
	freq_step = Fstep;
	//---------- Open-Loop Gain `H', Sensing Function `C', and ResponseFunction `RF'.
	modH = new double[Npoint];
	phH = new double[Npoint];
	modC = new double[Npoint];
	phC = new double[Npoint];
	modRF = new double[Npoint];
	phRF = new double[Npoint];
        //---------- Perform linear interpolation for H (last file read).
	LinearInterpolation(f,modH,Npoint,temp_f,temp_mod,temp_Npoint);
        std::cout << "# Finished LinearInterpolation() for modH.\n";
        LinearInterpolation(f,phH,Npoint,temp_f,temp_ph,temp_Npoint);
        std::cout << "# Finished LinearInterpolation() for phH.\n";
        delete[] temp_f;
        delete[] temp_mod;
        delete[] temp_ph;

	//---------- Re-read data from Sensing Function file.
        temp_Npoint = GetDataFileLength(sf_file);  
        temp_f = new double[temp_Npoint];
        temp_mod = new double[temp_Npoint];
        temp_ph = new double[temp_Npoint];
        ReadDataFile(sf_file,temp_Npoint,temp_f,temp_mod,temp_ph);
	//---------- Perform linear interpolation for C.
        LinearInterpolation(f,modC,Npoint,temp_f,temp_mod,temp_Npoint);
        std::cout << "# Finished LinearInterpolation() for modC.\n";
        LinearInterpolation(f,phC,Npoint,temp_f,temp_ph,temp_Npoint);
        std::cout << "# Finished LinearInterpolation() for phC.\n";
        //---------- Delete temporary variables.
        delete[] temp_f;
        delete[] temp_mod;
        delete[] temp_ph;
}


//Calibrate::Calibrate(const char *cal_file, const char *channel):TransFunc(NULL),dyncal(false),length(0),
//next_entry(0.0), last_entry(0.0), H1_ETMX(1.2), H2_ETMX(1.35), L1_ETMX(2.63),
//F_PEND(.75),ASQ_per_ADC(0.0),initial_amplitude(0.0), step(0), prev_end_time(0,0), var(0.0),
//CLOSENESS(.0025), UPPERBOUND(.05), LOWERBOUND(.02), gain_status(4), RECENT(5)

Calibrate::Calibrate(const char *cal_file, const char *channel):isalphagood(false),dyncal(false),
next_entry(0.0), last_entry(0.0), H1_ETMX(1.2), H2_ETMX(1.35), L1_ETMX(2.63),
F_PEND(.75), ASQ_per_ADC(0.0), CLOSENESS(.0025), UPPERBOUND(.05), 
LOWERBOUND(.02), initial_amplitude(0.0), var(0.0), TransFunc(NULL), length(0),
step(0), gain_status(4), RECENT(5), prev_end_time(0,0)
{
	//---------- Assign correct ETMX calibration (nm/ADCcount) based on 
        //           which interferometer the channel corresponds to.
	if(channel[0] == 'H') {
                if(channel[1] == '1')
                        ETMX = H1_ETMX;
                else
                        ETMX = H2_ETMX;
        }
        else
                ETMX = L1_ETMX;

	//---------- Zero last_few array
	last_few = new float[RECENT];
	for(int i = 0; i < RECENT; i++)
		last_few[i] = 0.0;

	//---------- Create file stream object to access data.
	std::ifstream input;
	input.open(cal_file);
	if(input.fail())
	{
		std::cerr << "SenseMonitor ERROR:  Failed to open calibration "
		 	  << "file " << cal_file << "!  Exiting.\n\n";
		std::cout << "SenseMonitor ERROR:  Failed to open calibration "
		 	  << "file " << cal_file << "!  Exiting.\n\n";
		exit(302);
	}

	//---------- Read through file once to determine necessary lengths 
        //           of arrays.  Read and discard column headers (assumed 
        //           to occupy the first two rows of the file).
	input.getline(line, 80);
	input.getline(line, 80); 
	while(input >> next_entry)
        {
                //---------- Test to determine if frequency is repeated.
                if(next_entry != last_entry)
                        length++;
                last_entry = next_entry;

                //---------- Read and discard the rest of the line.
                input.getline(line, 80);
        }
        
        //---------- Rezero for next trip through calibration file.
        last_entry = 0.0;
        next_entry = 0.0;

	//---------- Create arrays of appropriate length to hold freq and 
        //           module data.
	freq = new double[length];
	module = new double[length];

	//---------- Close and reopen file stream to return cursor to the 
        //           top of the file.
	input.close();
	std::ifstream input2(cal_file);
	if(input2.fail())
	{
		std::cerr << "SenseMonitor ERROR:  Failed to open calibration "
		 	  << "file " << cal_file << "!  Exiting.\n\n";
		std::cout << "SenseMonitor ERROR:  Failed to open calibration "
		 	  << "file " << cal_file << "!  Exiting.\n\n";
		exit(303);
	}

	//---------- Read and discard column headers, read calibration factors.
	input2.getline(line, 80);
	input2.getline(line, 80);
	int i = 0;
	while(input2 >> next_entry)
        {
                //---------- Ignore repeated frequencies.
                if(next_entry != last_entry)
                {
                        freq[i] = next_entry;
                        input2 >> module[i];
                        i++;
                }
                last_entry = next_entry;

                //---------- Read and discard rest of the line.
                input2.getline(line, 80);
        }

	//---------- Sort frequencies into increasing order.
	if (freq[0] > freq[length-1]) {
	    double *temp_freq = new double[length];
	    double *temp_module = new double[length];
	    for (int k=0; k<length; k++) {
                temp_freq[k] = freq[length-1-k];
                temp_module[k] = module[length-1-k];
	    }
	    for (int k=0; k<length; k++) {
                freq[k] = temp_freq[k];
                module[k] = temp_module[k];
	    }
	    delete[] temp_freq;
	    delete[] temp_module;
	}

        //---------- Record minimum and maximum frequencies available.
        Freq_min = freq[0];
        Freq_max = freq[length-1];
}
	

Calibrate::~Calibrate()
{
	//---------- Free up memory.
        if (!dyncal) {
	    delete[] freq;
	    delete[] module;
	    delete[] last_few;
	    delete TransFunc;
	} else {
            delete[] f;
            delete[] modH;
            delete[] phH;
            delete[] modC;
            delete[] phC;
            delete[] modRF;
            delete[] phRF;
        }
}


void Calibrate::DumpDynamicCalibration(std::ostream& out)
{
	out << "#\n";
	out << "# Interpolated data used in dynamic calibrations:\n";
	out << "# Magnitude modC is in (AS_Q counts)/(strain).\n";
	out << "#  Freq   modH   phH   modC   phC \n";
        // For default parameters using i+=60 prints out 
        // one data point every 10Hz.
	for (int i=0; i<Npoint; i+=60) {
	    out << "# " << f[i]; 
	    out << " " << modH[i] << " " << phH[i]; 
	    out << " " << modC[i] << " " << phC[i]; 
	    out << std::endl;
	}
	out << "# Finished interpolated data dump.\n";
	out << "#\n";
}
	

void Calibrate::DumpResponseFunction(std::ostream& out)
{
	out << "#\n";
	out << "# Snapshot response function used in dynamic calibrations:\n";
	out << "# Magnitude modRF is in (AS_Q counts)/(nanometer).\n";
	out << "#  Freq   modRF  phRF  \n";
        // For default parameters using i+=60 prints out 
        // one data point every 10Hz.
	for (int i=0; i<Npoint; i+=60) {
	    out << f[i]; 
	    out << " " << modRF[i] << " " << phRF[i]; 
	    out << std::endl;
	}
	out << "# Finished response function dump.\n";
	out << "#\n";
}
	

/*
Goal:
  Compute FSpectrum *ResponseFunc containing transfer function as calculated from 
  the open-loop gain, sensing function, and alpha, beta calibration parameters. 

Procedure:
  Response Function RF related to sensing function C, open-loop gain 
  H, and calibration parameters alpha and beta by

    RF = alpha*C/(1+alpha*beta*H) 

  where C = sf and H = olg are complex, and alpha, beta are real.  
  Assuming data files are in freq -- mod -- phase format, then 
  we want formulae for the modulus and phase of RF in terms of the 
  modulus and phase of C and H.

  Defining Hprime := alpha*beta*H, we have 

    ReHprime = alpha*beta*modH*cos(phH);
    ImHprime = alpha*beta*modH*sin(phH);

    1+Hprime = [1+alpha*beta*modH*cos(phH)] + i*[alpha*beta*modH*sin(phH)]
             = [1+ReHprime] + i*[ImHprime]
    mod(1+Hprime) = [(1+ReHprime)^2 +(ImHprime)^2]^0.5
    ph(1+Hprime) = atan([a*b*mH*sin(phH)]/[1+a*b*mH*cos(phH)])
             = atan(ImHprime/(1+ReHprime))

    modRF = alpha*modC/mod(1+Hprime)
          = alpha*modC/[(1+ReHprime)^2 +(ImHprime)^2]^0.5
    phRF = phC - ph(1+Hprime)
         = phC - atan(ImHprime/(1+ReHprime))

*/
//Note:  Units are AS_Qcounts/nanometer.
//:KLUDGE: Note that FSpectrum supports floats but not doubles - all of this precision may be unnecessary.
//void Calibrate::ComputeResponseFunction(double *modH, double *phH, double *modC, double *phC, int Npoint, double alpha, double beta)
void Calibrate::ComputeResponseFunction(void)
{
    //----- Make sure calibration parameters `alpha' and `beta' have been set.
    if ((alpha == 0.0) || (beta == 0.0)) {
	   std::cout << "SenseMonitor ERROR:  Calibrate::ComputeResponseFunction() "
	        << "called before Calibrate::setAlpha or Calibrate::setBeta().  "
		<< "Exiting.\n";
	   std::cerr << "SenseMonitor ERROR:  Calibrate::ComputeResponseFunction() "
		<< "called before Calibrate::setAlpha or Calibrate::setBeta().  "
		<< "Exiting.\n";
	   exit(79);
    }

    //---------- Temporary variables.
    float ReHprime;
    float ImHprime;
    for (int i=0; i<Npoint; ++i) { 
        ReHprime = alpha*beta*modH[i]*cos(phH[i]);
        ImHprime = alpha*beta*modH[i]*sin(phH[i]);
	//---------- Modulus of Response Function in AS_Q/strain.
        modRF[i] = pow(pow(1+ReHprime,2.0)+pow(ImHprime,2.0),-0.5);
        modRF[i] *= alpha*modC[i];
	//---------- Modulus of Response Function in AS_Q/nanometer.
	//           RF(ASQ/nm) = 1/L(nm)*RF(ASQ/strain).
        modRF[i] /= arm_length;
	//---------- Phase of Response Function.
        phRF[i] = phC[i] - atan(ImHprime/(1+ReHprime));
    }
}


// LinearInterpolation() assumes X,x are monotonically increasing (this is done for 
// data retrived by ReadDataFile()). 
void Calibrate::LinearInterpolation(double *X, double *Y, int N, double *x, double *y, int n)
{
    //----- Make sure that X ranges are subset of x.
    //if (X[0] < x[0]) std::cout << X[0] << " " << x[0] << " " << X[0]-x[0] << std::endl;
    //if (X[N-1] > x[n-1]) std::cout << X[N-1] << " " << x[n-1] << " " << X[N-1] - x[n-1] << std::endl;
    if ((X[0] < x[0]) || (X[N-1] > x[n-1])) {
        std::cerr << "SenseMonitor ERROR:  In Calibrate::LinearInterpolation() data "
             << "to be interpolated does not cover requested range.\n"
             << "Requested: [" << X[0] << "," << X[N-1] << "].\n"
             << "Available: [" << x[0] << "," << x[n-1] << "].\n"
             << "Exiting.\n";
        std::cout << "SenseMonitor ERROR:  In Calibrate::LinearInterpolation() data "
             << "to be interpolated does not cover requested range.\n"
             << "Requested: [" << X[0] << "," << X[N-1] << "].\n"
             << "Available: [" << x[0] << "," << x[n-1] << "].\n"
             << "Exiting.\n";
        exit(315);
    }

    //----- Perform interpolation.
    for (int i=0; i<N; ++i) { 
        //----- For each X[i] (0 <= i <= N-1) find int k 
        //      (0 <= k <= n-2) such that x[k] <= X[i] <= x[k+1].
        //      If no such k exists then data does not cover range of 
	//      interest, so exit with error message.
        int k=0; 
        while (!((x[k] <= X[i]) && (X[i] <= x[k+1]))) ++k;
        if (k>=n) {
            std::cerr << "SenseMonitor ERROR:  Calibration file data does not cover";
            std::cerr << " frequency range requested.\n";
	    std::cerr << "N = " << N << ", n = " << n << std::endl;
	    std::cerr << "X[" << i << "] = " << X[i] << ", x[" << k << "] = " << x[k] << std::endl; 
	    std::cerr << " Exiting.\n";
            std::cout << "SenseMonitor ERROR:  Calibration file data does not cover";
            std::cout << " frequency range requested.\n";
	    std::cout << "N = " << N << ", n = " << n << std::endl;
	    std::cout << "X[" << i << "] = " << X[i] << ", x[" << k << "] = " << x[k] << std::endl; 
	    std::cout << " Exiting.\n";
            exit(304);
        }
        //----- Interpolate for Y[i] using (x[k],y[k]), (x[k+1],y[k+1]).  
        double slope;
        double intercept;
        slope = (y[k+1] - y[k])/(x[k+1]-x[k]); 
        intercept = y[k] - slope*x[k];
        Y[i] = slope*X[i] + intercept;
    }
}


void Calibrate::InitializeTransferFunction(double f_min, double f_step, int nstep)
{
	double f_max = f_min + nstep*f_step;
        float *TransFunc_data = new float[nstep+1];

	//---------- Verify that calibration data covers frequency range of
	//           interest.  
	if ( (f_min < Freq_min) || (f_max > Freq_max) )  {
	    std::cerr << "SenseMonitor ERROR:  Calibration file data does not "
	         << "cover";
	    std::cerr << " frequency range requested.\n";
	    std::cerr << " Requested: [" << f_min << "," << f_max << "]Hz.\n";
	    std::cerr << " Available: [" << Freq_min << "," << Freq_max << "]Hz.\n";
	    std::cerr << " Exiting.\n";
	    std::cout << "SenseMonitor ERROR:  Calibration file data does not "
	         << "cover";
	    std::cout << " frequency range requested.\n";
	    std::cout << " Requested: [" << f_min << "," << f_max << "]Hz.\n";
	    std::cout << " Available: [" << Freq_min << "," << Freq_max << "]Hz.\n";
	    std::cout << " Exiting.\n";
  	    exit(305);
	}

	//---------- Initialize variables necessary for linear interpolation.
	double f_0 = f_min;
	double cal_factor = 0.0;
	double slope = 0.0, intercept = 0.0;
	int k = 0;

	//---------- Do linear interpolation. (This linear interpolation 
	//           works for both increasing and decreasing frequency.)
	for(int j = 0; j < nstep + 1; j++)
	{	
	        //---------- Do linear interpolation for calibration factor.
		while (cal_factor == 0.0) {
		    if (freq[k] == f_0) {
			cal_factor = module[k];
		    } else if (  ((freq[k]<f_0) && (f_0<freq[k+1]))  ||
				 ((freq[k]>f_0) && (f_0>freq[k+1]))  ) {
			slope = (module[k+1] - module[k])/(freq[k+1] - freq[k]);
			intercept = module[k] - slope * freq[k];
			cal_factor = slope * f_0 + intercept;
		    } else if (k >= length-1) {
			std::cerr << "SenseMonitor ERROR:  Calibration file data "
			     << "does not cover";
			std::cerr << " frequency range requested.  Exiting.\n";
			std::cout << "SenseMonitor ERROR:  Calibration file data "
			     << "does not cover";
			std::cout << " frequency range requested.  Exiting.\n";
			exit(306);
		    } else {
			++k;
		    }
		}
                TransFunc_data[j] = cal_factor;

		//----- Increment f_0.
		f_0 += f_step;
	
		//----- Reset cal_factor and k to 0.
		k = 0;
		cal_factor = 0.0;
	}
	
	//---------- Copy Transfer Function data into PSD.
	TransFunc = new FSpectrum(f_min, f_step, Time(0,0), Interval(0.0), nstep+1, TransFunc_data);
        delete[] TransFunc_data;
}


void Calibrate::ApplyDynamic(FSpectrum& Int_PSD)
{
        //:KLUDGE:WARNING:  Make sure you've called ComputeResponseFunction(void) first!
        //:KLUDGE:WARNING:  Extraction at f_Nyq may be screwed up.  Beware!
	//---------- Check that TransFunc has been initialized and 
	//           that frequencies match those of Int_PSD.
        //           See, eg, code in ApplyStatic();
	//:END:KLUDGE:

	//---------- Check that frequency resolution of interpolated 
	//           calibration data matches that of Fspectrum to be
	//           calibrated.
	if (freq_step != Int_PSD.getFStep()) {
            std::cerr << "SenseMonitor WARNING:  Frequency resolution of interpolated calibration " 
		 << "data does not match that of AS_Q FSpectrum.\n";
            std::cout << "SenseMonitor WARNING:  Frequency resolution of interpolated calibration " 
		 << "data does not match that of AS_Q FSpectrum.\n";
	}
	
        //---------- Extract subsection of FSpectrum for which we have calibration data.
        FSpectrum temp = Int_PSD.extract(freq_min, freq_max-freq_min+freq_step);
        Int_PSD = temp;
       
	//---------- Extract data sequences to float arrays.
	if (Int_PSD.getNStep() + 1 != Npoint) {
		std::cerr << "SenseMonitor ERROR:  Different number of points in calibration (" << Npoint << ") and AS_Q (" << Int_PSD.getNStep() + 1 << ") "
		     << "data.  Exiting.\n";
		std::cout << "SenseMonitor ERROR:  Different number of points in calibration (" << Npoint << ") and AS_Q (" << Int_PSD.getNStep() + 1 << ") "
		     << "data.  Exiting.\n";
		exit(307);
	}
        float *ASQ = new float[Npoint];
        Int_PSD.getData(Npoint, ASQ);
	
	//---------- Apply calibration.  Note that RF is in units of ASQ/nm.
        for (int i=0; i<Npoint; ++i) {
                ASQ[i] /= pow(modRF[i],2.0);
	}

	//---------- Get data back from array data and into Int_PSD
	Int_PSD.setData(Npoint, ASQ);

	//---------- Clean up.
        delete[] ASQ;
}


void Calibrate::ApplyStatic(FSpectrum& Int_PSD)
{
        //---------- Select subset of requested frequency range for which 
        //           we have calibration data.  Get low and high frequencies 
        //           as lowest/highest freq bin centers above/below freq 
        //           limits of PSD and calibration data.  Shave off extra 
        //           bin from each edge for safety.
        double f_step = Int_PSD.getFStep();
        freq_min = Int_PSD.getLowFreq(); 
        while (freq_min < Freq_min) {
            freq_min += f_step;
        }
        freq_min += f_step;
        freq_max = Int_PSD.getHighFreq(); 
        while (freq_max > Freq_max) {
            freq_max -= f_step;
        }
        freq_max -= f_step;

	//---------- Extract subset of Int_PSD for which we have 
        //           calibration data.
        //int nstep = ((int) (freq_max - freq_min)/f_step);
        FSpectrum temp = Int_PSD.extract(freq_min, freq_max - freq_min);
        Int_PSD = temp;
/*
        //:KLUDGE:WARNING:  Extraction at f_Nyq may be screwed up.  Beware!
        if (Int_PSD.getLowFreq() != freq_min) {
            std::cout << "WARNING: Int_PSD.getLowFreq() = " << Int_PSD.getLowFreq();
            std::cout << "   freq_min = " << freq_min << std::endl;
            freq_min = Int_PSD.getLowFreq();
        }
        if (Int_PSD.getHighFreq() != freq_max) {
            std::cout << "WARNING: Int_PSD.getHighFreq() = " << Int_PSD.getHighFreq();
            std::cout << "   freq_max = " << freq_max << std::endl;
            freq_max = Int_PSD.getHighFreq();
        }
        //:END:KLUDGE:
*/

	//---------- Check that TransFunc has been initialized and 
	//           that frequencies match those of Int_PSD.
        if (TransFunc == NULL) { 
	    //std::cout << "Calibrate MESSAGE:  Calling InitializeTransferFunction().\n";
	    InitializeTransferFunction(
		Int_PSD.getLowFreq(), Int_PSD.getFStep(), Int_PSD.getNStep()
	    );
	} else if ( (TransFunc->getLowFreq() != Int_PSD.getLowFreq()) ||
	            (TransFunc->getFStep() != Int_PSD.getFStep()) ||
	            (TransFunc->getNStep() != Int_PSD.getNStep()) ) { 
            std::cout << "Calibrate MESSAGE:  Mis-match in frequencies of FSpectrum ";
            std::cout << "to be calibrated and Transfer Function.  Calling ";
            std::cout << "InitializeTransferFunction().\n";
	    InitializeTransferFunction(
                Int_PSD.getLowFreq(), Int_PSD.getFStep(), Int_PSD.getNStep()
	    );
	}

	//---------- Extract data sequences to float arrays.
	int nsteps = Int_PSD.getNStep(); 
        float *psd_data = new float[nsteps+1];
        float *cal_data = new float[nsteps+1];
        Int_PSD.getData(nsteps + 1, psd_data);
        TransFunc->getData(nsteps + 1, cal_data);
	
	//---------- Combine data arrays with appropriate calibration factors.
	double f_0 = Int_PSD.getLowFreq();
	//double f_step = Int_PSD.getFStep();
	for(int j = 0; j < nsteps + 1; j++)
	{	
		//---------- Calibration formula:
		// PSD_{PhysMotion}(f) = PSD_{AS_Q}(f) * [ ETMX*(f_pend/f)^2 ]^2 /
		//                       [ TransFunc(f) ]^2
		psd_data[j] *= (pow((F_PEND / f_0),4) * pow(ETMX,2)) / 
			    pow(cal_data[j],2);

		//---------- Increment f_0
		f_0 += f_step;
	}
	
	//---------- Get data back from array data and into Int_PSD
	Int_PSD.setData(nsteps + 1, psd_data);

        delete[] psd_data;
        delete[] cal_data;
}


void Calibrate::ApplyStatic(FSpectrum& Int_PSD, double l_freq, double h_freq)
{
	//---------- Set up variables to step through Int_PSD and apply 
	//           calibration factors.
	FSpectrum temp = Int_PSD.extract(l_freq, (h_freq - l_freq) +
                                         Int_PSD.getFStep());
	int nsteps = temp.getNStep(); 

	float *data = new float[nsteps + 1];
	temp.getData(nsteps + 1, data);
	
	double f_0 = l_freq;
	double f_step = Int_PSD.getFStep();

	//---------- Verify that calibration data covers frequency range of
	// interest.  Note that frequencies can be increasing or decreasing.
        if (freq[0] < freq[length-1]) {
	    if ( (l_freq < freq[0]) || (h_freq > freq[length-1]) )  {
		std::cerr << "SenseMonitor ERROR:  Calibration file data does not cover";
		std::cerr << " frequency range requested.  Exiting.\n";
		std::cout << "SenseMonitor ERROR:  Calibration file data does not cover";
		std::cout << " frequency range requested.  Exiting.\n";
		exit(308);
	    }
	} else { 
	    if ( (l_freq < freq[length-1]) || (h_freq > freq[0]) )  {
		std::cerr << "SenseMonitor ERROR:  Calibration file data does not cover";
		std::cerr << " frequency range requested.  Exiting.\n";
		std::cout << "SenseMonitor ERROR:  Calibration file data does not cover";
		std::cout << " frequency range requested.  Exiting.\n";
		exit(309);
	    }
	} 
	
	//---------- Initialize variables nescessary for linear interpolation
	double cal_factor = 0.0;
	double slope = 0.0, intercept = 0.0;
	int k = 0;

	for(int j = 0; j < nsteps + 1; j++)
	{	
		// Search freq array for f_0.  Then get calibration factor
		// for that frequency.  Interpolate linearly if nescessary.
		// Note that frequencies can be increasing or decreasing.
		while (cal_factor == 0.0) {
		    if (freq[k] == f_0) {
			cal_factor = module[k];
		    } else if (  ((freq[k]<f_0) && (f_0<freq[k+1]))  ||
				 ((freq[k]>f_0) && (f_0>freq[k+1]))  ) {
			slope = (module[k+1] - module[k])/(freq[k+1] - freq[k]);
			intercept = module[k] - slope * freq[k];
			cal_factor = slope * f_0 + intercept;
		    } else if (k >= length-1) {
		        std::cerr << "SenseMonitor ERROR:  Calibration file data does not cover";
	 	        std::cerr << " frequency range requested.  Exiting.\n";
		        std::cout << "SenseMonitor ERROR:  Calibration file data does not cover";
 		        std::cout << " frequency range requested.  Exiting.\n";
			exit(310);
		    } else {
			++k;
		    }
		}
		// Multiply by the ratio of the pendulumn frequency to f_0 to 
		// the fourth power.
		// Multiply by the nm per ETMX count value squared.
		// Divide by the calculated cal_factor squared.
		data[j] *= (pow((F_PEND / f_0),4) * pow(ETMX,2)) / 
			    pow(cal_factor,2);

		// Increment f_0
		f_0 += f_step;
	
		// Reset cal_factor and k to 0
		k = 0;
		cal_factor = 0.0;
	}
	
	//---------- Get data back from array data and into Int_PSD
	temp.setData(nsteps + 1, data);
	Int_PSD = temp;

	delete[] data;
}


float Calibrate::getGain(TSeries *ts, LineFilter *Ref_Line,
	  		 const double lowerbound, const double upperbound,
			 const double closeness)
{
        //---------- Use LineFilter class object (Ref_Line) to track injected
	//           calibration line; rescale range if calibration has drifted.
        //           WARNING:  The measured power in the calibration line is
	//           sometimes very noisy, and tends to spike dramatically when 
	//           entering or leaving lock.  Don't use the amplitude for 
	//           calibration during these times or else we'll cause more 
	//           harm than good.  

	// Intialize LOWERBOUND and UPPERBOUND to user specified values
	LOWERBOUND = lowerbound;
	UPPERBOUND = upperbound;
	CLOSENESS = closeness * (upperbound - lowerbound);

	//----- Begin tracking amplitude of calibration line
	//Ref_Line->setFScan();  // Moved to SenseMonitor::SenseMonitor (S.K.)
        Ref_Line->apply(*ts);
        
	// Get amplitude from LineFilter object Ref_Line
	wavearray<float> temp;
        temp = Ref_Line->getTrend(0, 'a');
        float amplitude = temp[temp.size() - 1];
	
	//----- Check for a gap in the data (such as due to lock loss) 
	//      and if so, re-zero "last_few" array and initial_amplitude.
	if ( (ts->getStartTime()) != prev_end_time ) {
	    for(int i = 0; i < RECENT; i++) last_few[i] = 0.0;
	    initial_amplitude = 0.0;
	}
	prev_end_time = ts->getEndTime();
	
	// Manage dynamic parameters
	//float gain = 1.0;
	gain = 1.0;
	step = step % RECENT;
	last_few[step] = amplitude;
	step++;
	
	// Determine if any of the last RECENT number of calculated amplitudes
	// are out of the specified amplitude range.  If so, do not apply.
	for(int j = 0; j < RECENT; j++) {
        	if (last_few[j] <= LOWERBOUND) {  
			gain_status = 1;
			gain = 1.0;
                	return(gain);
                }
        	if (last_few[j] >= UPPERBOUND) {
			gain_status = 2;
			gain = 1.0;
                	return(gain);
                }
        }

	// Calculate the mean absolute difference between the consecutive 
	// last RECENT calculated amplitudes
	var = 0.0;
	for(int k = 0; k < RECENT - 1; k++)
        	var += fabs(last_few[((k+1+step) % RECENT )] - last_few[((k+step) % RECENT )]);
        var /= (float)(RECENT-1.0);
	
	// Determine whether data is good or not; if it is necessary, calculate
	// a new reference amplitude
	if(var > CLOSENESS) {
		gain_status = 3;
        	gain = 1.0;
		return(gain);
	} else if (initial_amplitude == 0.0) {
		for(int i = 0; i < RECENT; i++) {
			initial_amplitude += last_few[i];
		}
		initial_amplitude /= (double)(RECENT);
	}		

	// Calculate gain
	gain_status = 0;
	gain = amplitude / initial_amplitude;
        Ref_Line->resize(16);
	return((float) gain);
}


float Calibrate::getAmplitude(TSeries *ts, double Freq)
{
        //----- Estimate amplitude of calibration line using
        //      sum of power in bins around line freq.
        //      Procedure:
        //       1. Read in TSeries (preferably 100 sec).
        //       2. Compute PSD with Hanning window, 10 averages.
        //       3. Line power = power in bin containing line freq, 
        //          + power in neighboring bins (3 bins total).
        //          - noise power in counted bins,
        //            estimated as average power in nearby bins.
        //       4. Line ampl = [ 2*dF*Line power ]^(0.5)
        //          where dF = n_windows/T = frequency resolution.

        //----- Define needed objects.
        Hanning hann;
        PSD psd(&hann, 10);
        FSpectrum fs;
        psd.generate(fs,ts);
        double norm = pow( 2.0*(fs.getFStep()), 0.5);

        //----- Estimate line power using central bin plus one on either side:
        double power = fs.getSum(Freq - fs.getFStep(), 3.0*(fs.getFStep()) );

        //----- Estimate noise power/bin using 5 bins on either side:
        double noise = fs.getSum(Freq - 10.0*(fs.getFStep()), 5.0*(fs.getFStep()) );
        noise += fs.getSum(Freq + 5.0*(fs.getFStep()), 5.0*(fs.getFStep()) );
        noise /= 10.0;

        //----- Compute line amplitude in AS_Q counts.  Note that if line has 
	//      died we could get power - noise < 0, so check for that.
        double ampl;
        if (power > 3.0*noise) {
	    ampl = norm*pow(power-3.0*noise,0.5);
	} else {
	    ampl = 0.0;
	}
        //:KLUDGE:
        //std::cout << "Cal-line ampl = " << ampl << " ADC counts.\n";
        //:END:KLUDGE:

/*
        //:KLUDGE: Comment-out conversion to ADC counts.
	//----- Convert line amplitude to ADC counts.
	//      Search freq array for Freq, then get calibration factor
	//      for that frequency.  Interpolate linearly if necessary.
	//      Note that frequencies can be increasing or decreasing.
        //      NOTE: Should only be triggered first time called, since 
        //      after that will have ASQ_per_ADC != 0.0.
	double slope = 0.0, intercept = 0.0;
        int k = 0;
	while (ASQ_per_ADC == 0.0) {
	    if (freq[k] == Freq) {
		ASQ_per_ADC = module[k];
	    } else if (  ((freq[k]<Freq) && (Freq<freq[k+1]))  ||
			 ((freq[k]>Freq) && (Freq>freq[k+1]))  ) {
		slope = (module[k+1] - module[k])/(freq[k+1] - freq[k]);
		intercept = module[k] - slope * freq[k];
		ASQ_per_ADC = slope * Freq + intercept;
	    } else if (k >= length-1) {
		std::cout << "SenseMonitor ERROR:  Calibration file data does not cover";
		std::cout << " calibration-line frequency.  Exiting.\n";
		exit(311);
	    } else {
		++k;
	    }
            if (ASQ_per_ADC != 0.0) {
		std::cout << "TF ASQ ADC  at cal-line freq = " << ASQ_per_ADC << std::endl;
	    }
	}
	ampl /= ASQ_per_ADC;
	//:END:KLUDGE:
*/

        //----- Finished.
        return(ampl);
}


void Calibrate::ComputeAlpha(double freq, double ratio)
{
        //----- 
	// Procedure:  
	// 1. Get ReH, modH^2, mod(1+H)^2 at cal-line freq (do once).
	// 2. Get gain (usual old definition).
	// 3. Compute A, B, C (quadratic coeff).
	//      A = (gain)^2*(modH)^2 - mod(1+H)^2/beta^2
	//      B = 2*(gain)^2*ReH
	//      C = (gain)^2
	// 4. Solve 
	//      A(alpha*beta)^2 + B(alpha*beta) + C = 0.  
	//    Solution:
	//      alpha = (-B +/- [B^2-4*A*C]^0.5)/2*A*beta
	//    Tricky - have to look out for imaginary numbers.
	//    Temp:
	//      disc = B^2 - 4*A*C;
	//      if (disc < 0.0) std::cerr << "Help!\n";
	//      alpha1 = -B-pow(disc,0.5)/(2*A*beta);
	//      alpha2 = -B+pow(disc,0.5)/(2*A*beta);
	//    Expect alpha1 to be correct solution (this is the case 
	//    for (R->1,beta->1) and also for H->0.
	//    :KLUDGE:
	//      if ((alpha1 =< 0.0) && (alpha2 > 0.0)) {
	//          alpha = alpha2;
	//      } else if ((alpha1 > 0.0) && (alpha2 <= 0.0)) { 
	//          alpha = alpha1;
	//      } else { 
	//          std::cerr << "Bugger!\n";
	//      }
	if ((reh == 0.0) && (imh == 0.0)) {
	    double *X = new double[1];
	    double *Y = new double[1];
	    X[0] = freq;
	    LinearInterpolation(X, Y, 1, f, modH, Npoint); 
            double modh = Y[0];
	    LinearInterpolation(X, Y, 1, f, phH, Npoint); 
            double phh = Y[0];
	    reh = modh*cos(phh);  // compute these once, store forever!
	    imh = modh*sin(phh);
	    delete[] X;
	    delete[] Y;
	}
	double A = pow(ratio*reh,2.0) + pow(ratio*imh,2.0) - (pow(1+reh,2.0)+pow(imh,2.0))*pow(beta,-2.0);
	double B = 2.0*pow(ratio,2.0)*reh;
	double C = pow(ratio,2.0);

	//----- Compute alpha.
	//      (Must choose correct solution of quadratic equation.)
	double disc = pow(B,2.0) - 4.0*A*C;
	alpha = (-B-pow(disc,0.5))/(2*A*beta);

	//----- Verify that alpha is real and positive.
	if ((disc > 0.0) && (alpha > 0.0)) {
		isalphagood = true;
	} else {
		isalphagood = false;
		alpha = 1.0;
	}
/*
	if (disc <= 0.0) {
	    std::cerr << "SenseMonitor ERROR:  No real solution for calibration data.\n"
		 << "A = " << A << "  B = " << B << " C = " << C << std::endl
		 << "Exiting.\n";
	    std::cout << "SenseMonitor ERROR:  No real solution for calibration data.\n"
		 << "A = " << A << "  B = " << B << " C = " << C << std::endl
		 << "Exiting.\n";
            exit(316);
        }
	if (alpha <= 0.0) { 
	    std::cerr << "SenseMonitor ERROR:  alpha is non-positive.\n"
		 << "A = " << A << "  B = " << B << " C = " << C << std::endl
		 << "alpha = " << alpha << std::endl
		 << "Exiting.\n";
	    std::cout << "SenseMonitor ERROR:  alpha is non-positive.\n"
		 << "A = " << A << "  B = " << B << " C = " << C << std::endl
		 << "alpha = " << alpha << std::endl
		 << "Exiting.\n";
            exit(317);
	}
*/
}


void Calibrate::setAlpha(double value)
{
        alpha = value;
}


void Calibrate::setBeta(double value)
{
        beta = value;
}


void Calibrate::setGain(double value)
{
        gain = value;
}


double Calibrate::getAlpha(void)
{
	return alpha;
}


double Calibrate::getBeta(void)
{
	return beta;
}


double Calibrate::getGain(void)
{
	return gain;
}


/*
float Calibrate::getAmplitude(TSeries *ts, double Freq)
{
        //----- Estimate amplitude of calibration line
        //      using quadrature:
        //      ampl^2 = 2/N^2( [\int dt {sin(2pi ft)*tseries}]^2
        //                      + [\int dt {cos(2pi ft)*tseries}]^2 )
        double cossum = 0.0;
        double sinsum = 0.0;
        double phase = 0.0;
        double tstep = ts->getTStep();
        float tsdata[ ts->getNSample() ];
        ts->getData(ts->getNSample(),tsdata);
        for (int i=0; i < ts->getNSample(); i++) {
          phase = 2.0*3.141592654*Freq*((float) i)*tstep;
          cossum += (tsdata[i])*cos( phase );
          sinsum += (tsdata[i])*sin( phase );
        }
        double ampl = pow(pow(cossum,2.0)+pow(sinsum,2.0),0.5);
        ampl *= pow(2.0,0.5);
        ampl /= (ts->getNSample());
        return(ampl);
}
*/


float Calibrate::getAmplitude(TSeries *ts, LineFilter *Ref_Line)
{
	//----- Begin tracking amplitude of calibration line
        Ref_Line->setFScan();
        Ref_Line->apply(*ts);

        // Get amplitude from LineFilter object Ref_Line
        wavearray<float> temp;
        temp = Ref_Line->getTrend(0, 'a');
        float amplitude = temp[temp.size() - 1];

	return(amplitude);
}


int Calibrate::getGainStatus(void)
{
        //---------- Return value of calibration line amplitude status set
	//           in getGain() method.
	
	// Special state for recovering from lock loss.
	if (initial_amplitude == 0.0) 
		gain_status = 4;

	return gain_status;
}


double Calibrate::getLowFreq(void)  
{
	return Freq_min;
}


double Calibrate::getHighFreq(void)  
{
	return Freq_max;
}


/* 
This function reads a file and returns the number of rows.
It is assumed that each line begins with a double.  If two 
or more successive rows begin with the same number, rows 
after the first are ignored in counting (this is to handle 
a bug in a script that produces calibration files).
*/
int  
Calibrate::GetDataFileLength(const char *file)
{
        //---------- Define needed local variables.
        double prev_entry = -1.0;  //----- Previous 'column1' entry read from file.
        double curr_entry = -1.0;  //----- Current 'column1' entry read from file.
        int numb_entry = 0;       //----- Number of rows (not counted repeated column1 vaulues).
        char buffer[80];          //----- Holds unused data.

	//---------- Create file stream object to access data.
	std::ifstream input;
	input.open(file);
	if (input.fail()) {
		std::cout << "SenseMonitor ERROR:  Calibrate::GetDataFileLength() failed to open "
		 	  << "file " << file << ".  Exiting.\n\n";
		std::cerr << "SenseMonitor ERROR:  Calibrate::GetDataFileLength() failed to open "
		 	  << "file " << file << ".  Exiting.\n\n";
		exit(312);
	}

	//---------- Read through file once to determine length. 
        //           Ignore successive lines with the same first element.
	while (input >> curr_entry)
        {
                //---------- Test to determine if frequency is repeated.
                if (curr_entry != prev_entry)
                        numb_entry++;
                prev_entry = curr_entry;

                //---------- Read and discard the rest of the line.
                input.getline(buffer, 80);
        }
        
	//---------- Close file stream.
	input.close();

        return numb_entry;
}


void 
Calibrate::ReadDataFile(const char *file, int numb_entry, double *column1, double *column2, double *column3)
{
        //---------- Define needed local variables.
        double prev_entry = -1.0;  //----- Previous 'column1' entry read from file.
        double curr_entry = -1.0;  //----- Current 'column1' entry read from file.
        //int numb_entry = 0;       //----- Number of rows (not counted repeated column1 vaulues).
        char buffer[80];          //----- Holds unused data.

	//---------- Create and open file stream object to access data.
	std::ifstream input;
	input.open(file);
	if (input.fail()) {
		std::cout << "SenseMonitor ERROR:  Calibrate::ReadDataFile() failed to open "
		 	  << "file " << file << ".  Exiting.\n\n";
		std::cerr << "SenseMonitor ERROR:  Calibrate::ReadDataFile() failed to open "
		 	  << "file " << file << ".  Exiting.\n\n";
		exit(313);
	}

	//---------- Read data into arrays.
	int i = 0;
	while (input >> curr_entry) {
                //---------- Ignore lines with repeated value in column 1.
                if (curr_entry != prev_entry) {
                        column1[i] = curr_entry;
                        input >> column2[i];
                        input >> column3[i];
                        i++;
                }
                prev_entry = curr_entry;

                //---------- Read and discard rest of the line.
                //:KLUDGE: This may not be necessary, and may even be wrong.
                input.getline(buffer, 80);
        }

        //---------- Check that correct number of lines read.
        if (i != numb_entry) {
		std::cout << "SenseMonitor ERROR:  Calibrate::ReadDataFile() wrong number "
		 	  << "of lines in file " << file << ".  Found " << i << " and expected " << numb_entry << ".  Exiting.\n\n";
		std::cerr << "SenseMonitor ERROR:  Calibrate::ReadDataFile() found wrong number "
		 	  << "of lines in file " << file << ".  Exiting.\n\n";
		exit(314);
        }

	//---------- Sort data into increasing order by value of column1 data.
        //           Data is assumed already to be either monotonically 
        //           increasing or decreasing.
	if (column1[0] > column1[numb_entry-1]) {
	    double *temp_column1 = new double[numb_entry];
	    double *temp_column2 = new double[numb_entry];
	    double *temp_column3 = new double[numb_entry];
	    for (int k=0; k<numb_entry; k++) {
                temp_column1[k] = column1[numb_entry-1-k];
                temp_column2[k] = column2[numb_entry-1-k];
                temp_column3[k] = column3[numb_entry-1-k];
	    }
	    for (int k=0; k<numb_entry; k++) {
                column1[k] = temp_column1[k];
                column2[k] = temp_column2[k];
                column3[k] = temp_column3[k];
	    }
	    delete[] temp_column1;
	    delete[] temp_column2;
	    delete[] temp_column3;
	}

        //---------- Record minimum and maximum frequencies available.
        //Freq_min = freq[0];
        //Freq_max = freq[numb_entry-1];
}

//:KLUDGE: New cal code.	


//--marc start-----------------------------------------------------------------------

void Calibrate::ReadDataFile(const char *file, int numb_entry)
{
  /* 
   fstream InFile;
   double Num;
   int i = 0; //counter
   double* PSDArray = NULL;
   const int Max = numb_entry;
   

// Allocate storage for the array
   PSDArray = new double[Max];
  
// Open file
InFile.open("L1_CalPSD_GPS:752364960.txt", ios::in);
   if (InFile.fail())
      {
      cout << "Could not open  L1_CalPSD_GPS:752364960.txt --calibrate" << endl;
      exit(1);
      }     
 
   while(!InFile.eof() ) {
     InFile >> Num;
     PSDArray[i] = Num;
      	 if( PSDArray == NULL ) 
	 cout << "Failed to allocate memory" << endl;
     	i++;
          }
   // Close file 
  InFile.close(); 
  cerr << "THis line is only here to see if the program actually executed my function" << endl;
   
   // Release the memory allocate by new
   delete [] PSDArray;
   PSDArray = NULL;
  */
}
 
//-----------------------------------------------------------------------------marc end

void Calibrate::UpdateAlpha(TSeries* ts_asq, TSeries* ts_exc, double ampl_asq, double ampl_exc, double freq) 
{	
        //---------- Recompute alpha value using available data.
	gain = getAmplitude(ts_asq,freq)/getAmplitude(ts_exc,freq)*ampl_exc/ampl_asq;
	ComputeAlpha(freq,gain);
	return;
}


void Calibrate::UpdateBeta(TSeries** ts_darm, int num_darm_channels, double darm0)
{	
        //---------- Get darm average.  First try: multiply product of averages 
        //           of each channel (whole method asusmes beta does not change 
        //           much over one stride, so let's see if this works).
	beta = 1.0; 
        for (int j=0; j<num_darm_channels; ++j) {
	    beta *= ts_darm[j]->getAverage();
        }
	beta /= darm0;
	return;
}


bool Calibrate::IsCalibrationGood(void)
{
	return isalphagood;
}
//:END:KLUDGE:


#endif     //  Calibrate_HH


