#ifndef _EVENT_H_

#define _EVENT_H_

#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include <iostream>

#include <string>
#include <vector>
#include <queue>
#include <list>

#include <Time.hh>
#include <Interval.hh>
#include <FSpectrum.hh>

#include <errors.h>
#include <TidString.h>

typedef enum {Unknown, Diagnostic, Reference} EventType;

typedef enum {Undefined, Independent, Coincident} EventStatus;

typedef enum {Foetus, Child, Adult, Mature} EventMaturity;

/*

class TFCoords
{
 public:
  unsigned int t,  // time coordinate
    f;             // frequency coordinate
  float P;         // power

  TFCoords(){};
  ~TFCoords(){};

  bool operator < (const TFCoords& s); // to sort elements

  bool operator && (const TFCoords& s); // test if nearest neighbours
};

*/

//----------------------EVENT---------------------------------

#define EventCap 5 // default size of an event

class Event {  // internal event structure

  unsigned int cid; // channel Id

  EventType type; // event type

  EventMaturity maturity; // level of maturity of event

  Time time; // GPS start time of the event; by convention, the beginning of the interval used to compute the power

  Interval duration; // duration of event

  unsigned int size, // number of data points
    cap; // capacity

  unsigned int *t, *f; // time & frequency coordinates
  float *P; // power

  unsigned int minfreq, maxfreq;

 public:

  Event (); // constructor

  /*
  Event (unsigned int cid_,
	 const EventType type_,
	 const Time time_);
  */

  /*
  Event (unsigned int cid_,
	 const EventType type_,
	 const Time time_, 
	 const Interval duration_,
	 unsigned int t,
	 unsigned int f); // constructor
  */

  Event (unsigned int cid_,
	 const EventType type_,
	 const Time time_, 
	 const Interval duration_,
	 unsigned int t,
	 unsigned int f,
	 float P);

  //  Event (const Event &s); // copy constructor // use default (bitwise copy) copy constructor for efficiency

  unsigned int operator , (const Event& s); // get distance between events

  bool operator < (const Event& s); // to sort elements by LastTimeID

  void operator += (const Event& b); // merge two events

  bool operator && (const Event& b); // test if events touch

  unsigned int LastTimeID () {return t[size-1];} // return oldest pixel

  unsigned int FirstTimeID () {return *t;} // return oldest pixel

  void BeAdult() {maturity = Adult;} // set to Adult

  void BeMature() {maturity = Mature;} // set to mature

  TidString* Marshal(); // convert info to string

  EventMaturity GetMaturity() {return maturity;} 

  EventType GetType () {return type;}

  unsigned int GetSize() {return size;}

  unsigned int GetChannelID () {return cid;}

  float Power (); // compute total power

  float ExcessPower (vector<float>& PThresh); // compute total excess power

  void Reset (); // reset values to those of default constructor

  void Set(unsigned int cid_,
	   const EventType type_,
	   const Time time_, 
	   const Interval duration_,
	   unsigned int t,
	   unsigned int f,
	   float P);  // set all values

  void IncreaseCap (); // increase cap to 2*size

  unsigned int GetMaxFreq () {return maxfreq;}

  unsigned int GetMinFreq () {return minfreq;}

  ~Event (); // destructor

  void AppendEventToFile (char* filename); // debugging crap
};


void swrite(char *, void *, size_t); // convert data to byte array

void SWrite(TidString* s, void *ptr, size_t siz);

//------------------EVENT LIST------------------------
class EventsList {

  bool Raw; // set to 1 to get no clustering analysis or other kind of thresholding

  vector<float> PThresh; // threshold on power, goes from DC to Nyquist

  float alpha; // confidence level for power test

  unsigned int sigma; // minimum cluster size

  unsigned int **delta; // array of distances delta, with sigma-1 rows, row i having i columns

  unsigned int maxDt; // max of delta
  unsigned int minSizeDelta; // min size with non-zero delta

  unsigned int maxClusterDuration; // maximum number of timesteps a cluster can have

  string chnname; // channel name

  EventType type; // type of events in this list

  double fLow, // minimum frequency (Hz)
    fHigh, // maximum freq
    fRes;  // frequency resolution (bandwidth)
  bool useNyquist; // set to 1 so that fHigh is always the Nyquist frequency

  Time firstTime; // time of beginning of event list 

  Interval deltaT; // duration of a stretch of data used to compute the power


  list<Event*> eventsList; // list of events being processed

  unsigned int nowID; // time Id of last event


  void Cluster (); // clean the list by merging and deleting, i.e., do nearest neighbour clustering analysis

  void CThresh (); // clean the list by applying clusters thresholds

  void CPower (); // apply power threshold on mature events

 public:

  EventsList (); // constructor

  EventsList (string chnname_,
	      unsigned int sigma_,
	      vector<unsigned int>& delta_,
	      float alpha_,
	      vector<float>& pthresh_,
	      EventType type_,
	      double fLow_,
	      double fHigh_,   // set to <0 to get Nyquist
	      Interval deltaT_); // constructor

  /*
  EventsList (string chnname_,
	      EventType type_,
	      double fLow_,
	      double fHigh_,
	      Interval deltaT_); // constructor
  */

  void Add(Event* s); // add new event at end of list

  void Add(unsigned int cid,
	   const FSpectrum* s); // add all events from a given Spectrum object

  void Clean ();

  void Purge(); // remove Mature events and send them out

  list<Event*>::iterator Erase(list<Event*>::iterator); // remove on element and free Event obj; return iterator of object *after* the one deleted

  void SetnowID(unsigned int s) {nowID = s;}

  bool EventsList::CloseEnough(Event *p, Event *q); // apply delta test

  void SetDelta (const unsigned int sigma_, 
		 const unsigned int* deltaVect); // set parameters 

  void SetDelta (const unsigned int sigma_, 
		 vector<unsigned int>& deltaVect); // set parameters 

  void SetAlpha (const float alpha);

  void SetPThresh (const vector<float>& pthresh);

  void Dump ();

  void MatureAll();

  ~EventsList (); // destructor

  void Clear(); // remove all events from eventsList

  // inline functions
  void SetPThresh (vector<float>& pthresh_) { PThresh = pthresh_; }
  void IncrementNowId () { nowID++; }
  void SetNow (unsigned int n) { nowID = n; }
  void BeRaw () {Raw = 1;}
  string GetName () {return chnname;}
  unsigned int GetSigma () {return sigma;}
  unsigned int GetDelta (unsigned int i, unsigned j) {return delta[i][j];}
  float GetAlpha () {return alpha;}
  //  float GetPThresh () {return PThresh;}
  bool GetuseNyquist () {return useNyquist;}
  double GetfLow () {return fLow;}
  double GetfHigh () {return fHigh;}
  unsigned int GetNFreq () {return (unsigned int)floor((fHigh - fLow)/fRes);}
  void SetMaxClusterDuration (unsigned int a) { maxClusterDuration = a; }

  void SetNowId(Time);
};


#endif
